# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.server

Handles all communication with Unity through a Flask service.  Messages
are sent to the controller using a pair of request/response queues.
"""

import warnings
import numpy as np
from enum import Enum
from ai2thor.util.depth import apply_real_noise, generate_noise_indices
import json
import sys


class NumpyAwareEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return obj.item()
        return super(NumpyAwareEncoder, self).default(obj)


class MultiAgentEvent(object):
    def __init__(self, active_agent_id, events):
        self._active_event = events[active_agent_id]
        self.metadata = self._active_event.metadata
        self.screen_width = self._active_event.screen_width
        self.screen_height = self._active_event.screen_height
        self.events = events
        self.third_party_camera_frames = []
        # XXX add methods for depth,sem_seg

    def __bool__(self):
        return bool(self._active_event)

    @property
    def cv2img(self):
        return self._active_event.cv2img

    def add_third_party_camera_image(self, third_party_image_data):
        self.third_party_camera_frames.append(
            read_buffer_image(
                third_party_image_data, self.screen_width, self.screen_height
            )
        )


def read_buffer_image(
    buf, width, height, flip_y=True, flip_x=False, dtype=np.uint8, flip_rb_colors=False
):
    im_bytes = (
        np.frombuffer(buf.tobytes(), dtype=dtype)
        if sys.version_info.major < 3
        else np.frombuffer(buf, dtype=dtype)
    )
    im = im_bytes.reshape(height, width, -1)
    if flip_y:
        im = np.flip(im, axis=0)
    if flip_x:
        im = np.flip(im, axis=1)
    if flip_rb_colors:
        im = im[..., ::-1]

    return im


def unique_rows(arr, return_index=False, return_inverse=False):
    arr = np.ascontiguousarray(arr).copy()
    b = arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[1])))
    if return_inverse:
        _, idx, inv = np.unique(b, return_index=True, return_inverse=True)
    else:
        _, idx = np.unique(b, return_index=True)
    unique = arr[idx]
    if return_index and return_inverse:
        return unique, idx, inv
    elif return_index:
        return unique, idx
    elif return_inverse:
        return unique, inv
    else:
        return unique


class MetadataWrapper(dict):
    def __getitem__(self, x):
        # alias deprecated functionality
        if x == "reachablePositions":
            last_action = super().__getitem__("lastAction")
            if last_action == "GetReachablePositions":
                warnings.warn(
                    'The key event.metadata["reachablePositions"] is deprecated and has been remapped to event.metadata["actionReturn"].'
                )
                x = "actionReturn"
            elif last_action == "GetSceneBounds":
                # Undocumented GetSceneBounds used to only populate reachablePositions,
                # and not actionReturn. This now maintains both sideways and
                # backwards compatibility in such a case.
                if "reachablePositions" in self:
                    return super().__getitem__(x)
                else:
                    warnings.warn(
                        'The key event.metadata["reachablePositions"] is deprecated and has been remapped to event.metadata["actionReturn"].'
                    )
                    x = "actionReturn"
            else:
                raise IndexError(
                    "You are trying to access event.metadata['reachablePositions'] without first "
                    + "calling controller.step(action='GetReachablePositions'). Also, "
                    + "the key 'reachablePositions' is deprecated in favor of event.metadata['actionReturn']."
                )
        return super().__getitem__(x)


class Event:
    """
    Object that is returned from a call to controller.step().
    This class wraps the screenshot that Unity captures as well
    as the metadata sent about each object
    """

    def __init__(self, metadata):
        self.metadata = MetadataWrapper(metadata)
        self.screen_width = metadata["screenWidth"]
        self.screen_height = metadata["screenHeight"]

        self.frame = None
        self.depth_frame = None
        self.normals_frame = None
        self.flow_frame = None

        self.color_to_object_id = {}
        self.object_id_to_color = {}

        self.instance_detections2D = None
        self.instance_masks = {}
        self.class_masks = {}

        self.instance_segmentation_frame = None
        self.semantic_segmentation_frame = None

        self.class_detections2D = {}

        self.process_colors()
        self.process_visible_bounds2D()
        self.third_party_camera_frames = []
        self.third_party_semantic_segmentation_frames = []
        self.third_party_instance_segmentation_frames = []
        self.third_party_depth_frames = []
        self.third_party_normals_frames = []
        self.third_party_flows_frames = []

        self.events = [self]  # Ensure we have a similar API to MultiAgentEvent

    def __bool__(self):
        return self.metadata["lastActionSuccess"]

    def __repr__(self):
        """Summarizes the results from an Event."""
        action_return = str(self.metadata["actionReturn"])
        if len(action_return) > 100:
            action_return = action_return[:100] + "..."
        return (
            "<ai2thor.server.Event at "
            + str(hex(id(self)))
            + '\n    .metadata["lastAction"] = '
            + str(self.metadata["lastAction"])
            + '\n    .metadata["lastActionSuccess"] = '
            + str(self.metadata["lastActionSuccess"])
            + '\n    .metadata["errorMessage"] = "'
            + str(self.metadata["errorMessage"]).replace("\n", " ")
            + '\n    .metadata["actionReturn"] = '
            + action_return
            + "\n>"
        )

    def __str__(self):
        return self.__repr__()

    @property
    def image_data(self):
        warnings.warn(
            "Event.image_data has been removed - RGB data can be retrieved from event.frame and encoded to an image format"
        )
        return None

    @property
    def class_segmentation_frame(self):
        warnings.warn(
            "event.class_segmentation_frame has been renamed to event.semantic_segmentation_frame.",
            DeprecationWarning,
        )
        return self.semantic_segmentation_frame

    def process_visible_bounds2D(self):
        if self.instance_detections2D and len(self.instance_detections2D) > 0:
            for obj in self.metadata["objects"]:
                obj["visibleBounds2D"] = (
                    obj["visible"] and obj["objectId"] in self.instance_detections2D
                )

    def process_colors(self):
        if "colors" in self.metadata and self.metadata["colors"]:
            for color_data in self.metadata["colors"]:
                name = color_data["name"]
                c_key = tuple(color_data["color"])
                self.color_to_object_id[c_key] = name
                self.object_id_to_color[name] = c_key

    def objects_by_type(self, object_type):
        return [
            obj for obj in self.metadata["objects"] if obj["objectType"] == object_type
        ]

    def process_colors_ids(self):
        if self.instance_segmentation_frame is None:
            return

        MIN_DETECTION_LEN = 0

        self.instance_detections2D = {}
        unique_ids, unique_inverse = unique_rows(
            self.instance_segmentation_frame.reshape(-1, 3), return_inverse=True
        )
        unique_inverse = unique_inverse.reshape(
            self.instance_segmentation_frame.shape[:2]
        )
        unique_masks = (
            np.tile(unique_inverse[np.newaxis, :, :], (len(unique_ids), 1, 1))
            == np.arange(len(unique_ids))[:, np.newaxis, np.newaxis]
        )
        # for unique_color_ind, unique_color in enumerate(unique_ids):
        for color_bounds in self.metadata["colorBounds"]:
            color = np.array(color_bounds["color"])
            color_name = self.color_to_object_id.get(
                tuple(int(cc) for cc in color), "background"
            )
            cls = color_name
            simObj = False
            if "|" in cls:
                cls = cls.split("|")[0]
                simObj = True

            bb = np.array(color_bounds["bounds"])
            bb[[1, 3]] = self.metadata["screenHeight"] - bb[[3, 1]]
            if not (
                (bb[2] - bb[0]) < MIN_DETECTION_LEN
                or (bb[3] - bb[1]) < MIN_DETECTION_LEN
            ):
                if cls not in self.class_detections2D:
                    self.class_detections2D[cls] = []

                self.class_detections2D[cls].append(bb)

                color_ind = np.argmin(np.sum(np.abs(unique_ids - color), axis=1))

                if simObj:
                    self.instance_detections2D[color_name] = bb
                    self.instance_masks[color_name] = unique_masks[color_ind, ...]

                if cls not in self.class_masks:
                    self.class_masks[cls] = unique_masks[color_ind, ...]
                else:
                    self.class_masks[cls] = np.logical_or(
                        self.class_masks[cls], unique_masks[color_ind, ...]
                    )

    def _image_depth(self, image_depth_data, **kwargs):
        image_depth = read_buffer_image(
            image_depth_data, self.screen_width, self.screen_height
        )
        depth_format = kwargs["depth_format"]
        image_depth_out = (
            image_depth[:, :, 0]
            + image_depth[:, :, 1] / np.float32(256)
            + image_depth[:, :, 2] / np.float32(256 ** 2)
        )
        multiplier = 1.0
        if depth_format != DepthFormat.Normalized:
            multiplier = kwargs["camera_far_plane"] - kwargs["camera_near_plane"]
        elif depth_format == DepthFormat.Millimeters:
            multiplier *= 1000
        image_depth_out *= multiplier / 256.0

        depth_image_float = image_depth_out.astype(np.float32)

        if "add_noise" in kwargs and kwargs["add_noise"]:
            depth_image_float = apply_real_noise(
                depth_image_float, self.screen_width, indices=kwargs["noise_indices"]
            )

        return depth_image_float

    def add_image_depth_robot(self, image_depth_data, depth_format, **kwargs):
        multiplier = 1.0
        camera_far_plane = kwargs.pop("camera_far_plane", 1)
        camera_near_plane = kwargs.pop("camera_near_plane", 0)
        if depth_format == DepthFormat.Normalized:
            multiplier = 1.0 / (camera_far_plane - camera_near_plane)
        elif depth_format == DepthFormat.Millimeters:
            multiplier = 1000.0

        image_depth = (
            read_buffer_image(
                image_depth_data, self.screen_width, self.screen_height, **kwargs
            ).reshape(self.screen_height, self.screen_width)
            * multiplier
        )
        self.depth_frame = image_depth.astype(np.float32)

    def add_image_depth(self, image_depth_data, **kwargs):
        self.depth_frame = self._image_depth(image_depth_data, **kwargs)

    def add_third_party_image_depth(self, image_depth_data, **kwargs):
        self.third_party_depth_frames.append(
            self._image_depth(image_depth_data, **kwargs)
        )

    def add_third_party_image_normals(self, normals_data):
        self.third_party_normals_frames.append(
            read_buffer_image(normals_data, self.screen_width, self.screen_height)
        )

    def add_image_normals(self, image_normals_data):
        self.normals_frame = read_buffer_image(
            image_normals_data, self.screen_width, self.screen_height
        )

    def add_third_party_image_flows(self, flows_data):
        self.third_party_flows_frames.append(
            read_buffer_image(flows_data, self.screen_width, self.screen_height)
        )

    def add_image_flows(self, image_flows_data):
        self.flows_frame = read_buffer_image(
            image_flows_data, self.screen_width, self.screen_height
        )

    def add_third_party_camera_image(self, third_party_image_data):
        self.third_party_camera_frames.append(
            read_buffer_image(
                third_party_image_data, self.screen_width, self.screen_height
            )
        )

    def add_image(self, image_data, **kwargs):
        self.frame = read_buffer_image(
            image_data, self.screen_width, self.screen_height, **kwargs
        )

    def add_image_ids(self, image_ids_data):
        self.instance_segmentation_frame = read_buffer_image(
            image_ids_data, self.screen_width, self.screen_height
        )
        self.process_colors_ids()

    def add_third_party_image_ids(self, image_ids_data):
        self.third_party_instance_segmentation_frames.append(
            read_buffer_image(image_ids_data, self.screen_width, self.screen_height)
        )

    def add_image_classes(self, image_classes_data):
        self.semantic_segmentation_frame = read_buffer_image(
            image_classes_data, self.screen_width, self.screen_height
        )

    def add_third_party_image_classes(self, image_classes_data):
        self.third_party_semantic_segmentation_frames.append(
            read_buffer_image(image_classes_data, self.screen_width, self.screen_height)
        )

    def cv2image(self):
        warnings.warn("Deprecated - please use event.cv2img")
        return self.cv2img

    @property
    def cv2img(self):
        return self.frame[..., ::-1]

    @property
    def pose(self):
        agent_meta = self.metadata["agent"]
        loc = agent_meta["position"]
        rotation = round(agent_meta["rotation"]["y"] * 1000)
        horizon = round(agent_meta["cameraHorizon"] * 1000)
        return (round(loc["x"] * 1000), round(loc["z"] * 1000), rotation, horizon)

    @property
    def pose_discrete(self):
        # XXX should have this as a parameter
        step_size = 0.25
        agent_meta = self.metadata["agent"]
        loc = agent_meta["position"]
        rotation = int(agent_meta["rotation"]["y"] / 90.0)
        horizon = int(round(agent_meta["cameraHorizon"]))
        return (int(loc["x"] / step_size), int(loc["z"] / step_size), rotation, horizon)

    def get_object(self, object_id):
        for obj in self.metadata["objects"]:
            if obj["objectId"] == object_id:
                return obj
        return None


class DepthFormat(Enum):
    Meters = (0,)
    Normalized = (1,)
    Millimeters = 2


class Server(object):
    def __init__(
        self, width, height, depth_format=DepthFormat.Meters, add_depth_noise=False
    ):
        self.depth_format = depth_format
        self.add_depth_noise = add_depth_noise
        self.noise_indices = None
        self.camera_near_plane = 0.1
        self.camera_far_plane = 20.0
        self.sequence_id = 0
        self.started = False
        self.client_token = None
        self.unity_proc = None

        if add_depth_noise:
            assert width == height, "Noise supported with square dimension images only."
            self.noise_indices = generate_noise_indices(width)

    def set_init_params(self, init_params):
        self.camera_near_plane = init_params["cameraNearPlane"]
        self.camera_far_plane = init_params["cameraFarPlane"]

    def create_event(self, metadata, files):
        if metadata["sequenceId"] != self.sequence_id:
            raise ValueError(
                "Sequence id mismatch: %s vs %s"
                % (metadata["sequenceId"], self.sequence_id)
            )

        events = []

        for i, a in enumerate(metadata["agents"]):
            e = Event(a)
            image_mapping = dict(
                image=e.add_image,
                image_depth=lambda x: e.add_image_depth(
                    x,
                    depth_format=self.depth_format,
                    camera_near_plane=self.camera_near_plane,
                    camera_far_plane=self.camera_far_plane,
                    add_noise=self.add_depth_noise,
                    noise_indices=self.noise_indices,
                ),
                image_ids=e.add_image_ids,
                image_classes=e.add_image_classes,
                image_normals=e.add_image_normals,
                image_flow=e.add_image_flows,
            )

            for key in image_mapping.keys():
                if key in files:
                    image_mapping[key](files[key][i])

            third_party_image_mapping = {
                # if we want to convert this param to underscores in Unity, we will need to
                # keep the mapping with the dash for backwards compatibility with older
                # Unity builds
                "image-thirdParty-camera": e.add_third_party_camera_image,
                "image_thirdParty_depth": lambda x: e.add_third_party_image_depth(
                    x,
                    depth_format=self.depth_format,
                    camera_near_plane=self.camera_near_plane,
                    camera_far_plane=self.camera_far_plane,
                ),
                "image_thirdParty_image_ids": e.add_third_party_image_ids,
                "image_thirdParty_classes": e.add_third_party_image_classes,
                "image_thirdParty_normals": e.add_third_party_image_normals,
                "image_thirdParty_flows": e.add_third_party_image_flows,
            }

            if a["thirdPartyCameras"] is not None:
                for ti, t in enumerate(a["thirdPartyCameras"]):
                    for key in third_party_image_mapping.keys():
                        if key in files:
                            third_party_image_mapping[key](files[key][ti])
            events.append(e)

        if len(events) > 1:
            self.last_event = MultiAgentEvent(metadata["activeAgentId"], events)
        else:
            self.last_event = events[0]

        return self.last_event



import os
import random
import time


def makedirs(directory):
    os.makedirs(directory, exist_ok=True)


def atomic_write(path, data):
    tmp_path = "-".join([path, str(time.time()), str(random.random())])
    mode = "w"

    if type(data) is bytes:
        mode = "wb"

    with open(tmp_path, mode) as f:
        f.write(data)
    os.rename(tmp_path, path)



import json
import math
import copy


def vector_distance(v0, v1):
    dx = v0["x"] - v1["x"]
    dy = v0["y"] - v1["y"]
    dz = v0["z"] - v1["z"]
    return math.sqrt(dx * dx + dy * dy + dz * dz)


def path_distance(path):
    distance = 0
    for i in range(0, len(path) - 1):
        distance += vector_distance(path[i], path[i + 1])
    return distance


def compute_spl(episodes_with_golden):
    """
    Computes batch SPL from episode list
    :param episodes_with_golden:
        sequence of episode object, an episode should include the following keys:
                                 'path'
                                    the path to evaluate and 'shortest_path' the shortest path
                                    as returned by 'get_shortest_path_to_object'.
                                    Both as sequences with values of the form
                                    dict(x=float, y=float, z=float)
                                 'success' boolean, a 0 for a failed path 1 for a successful one
    :return: returns a float representing the spl
    """
    N = len(episodes_with_golden)
    eval_sum = 0.0
    for i, episode in enumerate(episodes_with_golden):
        path = episode["path"]
        shortest_path = episode["shortest_path"]
        eval_sum += compute_single_spl(path, shortest_path, episode["success"])
    return eval_sum / N


def compute_single_spl(path, shortest_path, successful_path):
    """
    Computes SPL for a path dict(x=float, y=float, z=float)
    :param path: Sequence of dict(x=float, y=float, z=float) representing the path to evaluate
    :param shortest_path: Sequence of dict(x=float, y=float, z=float) representing the shortest oath
    :param successful_path: boolean indicating if the path was successful, 0 for a failed path or 1 for a successful one
    :return:
    """
    Si = 1 if successful_path == True or successful_path == 1 else 0
    li = path_distance(shortest_path)
    pi = path_distance(path)
    if max(pi, li) > 0:
        pl_ratio = li / max(pi, li)
    else:
        pl_ratio = 1.0
    spl = Si * pl_ratio
    return spl


def get_shortest_path_to_object(
    controller, object_id, initial_position, initial_rotation=None
):
    """
    Computes the shortest path to an object from an initial position using a controller
    :param controller: agent controller
    :param object_id: string with id of the object
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param initial_rotation: dict(x=float, y=float, z=float) representing rotation around axes or None
    :return:
    """
    args = dict(
        action="GetShortestPath",
        objectId=object_id,
        position=initial_position,
    )
    if initial_rotation is not None:
        args["rotation"] = initial_rotation
    event = controller.step(args)
    if event.metadata["lastActionSuccess"]:
        return event.metadata["actionReturn"]["corners"]
    else:
        raise ValueError(
            "Unable to find shortest path for objectId '{}'".format(object_id)
        )


def get_shortest_path_to_object_type(
    controller, object_type, initial_position, initial_rotation=None, allowed_error=None
):
    """
    Computes the shortest path to an object from an initial position using a controller
    :param controller: agent controller
    :param object_type: string that represents the type of the object
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param initial_rotation: dict(x=float, y=float, z=float) representing rotation around axes or None
    :param allowed_error: a floating point number describing the total amount of error allowed in specifying
        the start and end point in the shortest path computation. This number should be non-zero to allow for
        floating point issues and can be made larger to make this method more robust to edge cases where
        ai2thor "thinks" no path exists (this comes at the cost of some added noise to the start/end positions of the
        path). Passing `None` to this argument (the default) will result in THOR choosing it to be some small value,
        note that this value will not be exactly 0 to be robust to floating point inaccuracies.

    """
    kwargs = dict(
        action="GetShortestPath",
        objectType=object_type,
        position=initial_position,
    )
    if initial_rotation is not None:
        kwargs["rotation"] = initial_rotation
    if allowed_error is not None:
        kwargs["allowedError"] = allowed_error

    event = controller.step(kwargs)
    if event.metadata["lastActionSuccess"]:
        return event.metadata["actionReturn"]["corners"]
    else:
        raise ValueError(
            "Unable to find shortest path for object type '{}' due to error '{}'.".format(
                object_type, event.metadata["errorMessage"]
            )
        )


def get_shortest_path_to_point(
    controller, initial_position, target_position, allowed_error=None
):
    """
    Computes the shortest path to a point from an initial position using an agent controller
    :param controller: agent controller
    :param initial_position: dict(x=float, y=float, z=float) with the desired initial rotation
    :param target_position: dict(x=float, y=float, z=float) representing target position
    :param allowed_error: See documentation of the `get_shortest_path_to_object_type` method.
    :return:
    """
    kwargs = dict(
        action="GetShortestPathToPoint",
        position=initial_position,
        x=target_position["x"],
        y=target_position["y"],
        z=target_position["z"],
    )
    if allowed_error is not None:
        kwargs["allowedError"] = allowed_error

    event = controller.step(kwargs)
    if event.metadata["lastActionSuccess"]:
        return event.metadata["actionReturn"]["corners"]
    else:
        raise ValueError(
            "Unable to find shortest path to point '{}'  due to error '{}'.".format(
                target_position, event.metadata["errorMessage"]
            )
        )


def get_episodes_with_shortest_paths(controller, episodes):
    """
    Computes shortest path for an episode sequence
    :param controller: agent controller
    :param episodes: sequence of episode object required fields:
                        'target_object_id' string representing the object to look for
                        'initial_position' dict(x=float, y=float, z=float) of starting position
                        'initial_rotation' dict(x=float, y=float, z=float) representing rotation
                                           around axes
    :return:
    """
    episodes_with_golden = copy.deepcopy(episodes)
    for _, episode in enumerate(episodes_with_golden):
        controller.reset(episode["scene"])

        try:
            if "target_object_id" in episode:
                episode["shortest_path"] = get_shortest_path_to_object(
                    controller,
                    episode["target_object_id"],
                    {
                        "x": episode["initial_position"]["x"],
                        "y": episode["initial_position"]["y"],
                        "z": episode["initial_position"]["z"],
                    },
                    {
                        "x": episode["initial_rotation"]["x"],
                        "y": episode["initial_rotation"]["y"],
                        "z": episode["initial_rotation"]["z"],
                    },
                )
            else:
                episode["shortest_path"] = get_shortest_path_to_object_type(
                    controller,
                    episode["target_object_type"],
                    {
                        "x": episode["initial_position"]["x"],
                        "y": episode["initial_position"]["y"],
                        "z": episode["initial_position"]["z"],
                    },
                    {
                        "x": episode["initial_rotation"]["x"],
                        "y": episode["initial_rotation"]["y"],
                        "z": episode["initial_rotation"]["z"],
                    },
                )
        except ValueError:
            raise ValueError(
                "Unable to find shortest path for objectId '{}' in episode '{}'".format(
                    episode["target_object_id"],
                    json.dumps(episode, sort_keys=True, indent=4),
                )
            )
    return episodes_with_golden



import copy

import glob
import shutil
import json
import pickle
import ai2thor.controller
from collections import defaultdict
import os

MOVE_AHEAD_MAP = {
    0: dict(x=0, z=1),
    90: dict(x=1, z=0),
    180: dict(x=0, z=-1),
    270: dict(x=-1, z=0),
}


class Controller(object):
    def __init__(self, base_dir, grid_size=0.25):

        self.grid_size = grid_size
        self.base_dir = base_dir
        if grid_size < 0.25:
            raise Exception(
                "must adjust granularity of key_for_point for smaller grid sizes"
            )

    def reset(self, scene_name):
        self.scene_name = scene_name
        with open("%s/%s/index.json" % (self.base_dir, self.scene_name)) as f:
            self.positions = json.loads(f.read())

        # for g in glob.glob('%s/%s/metadata/*.json' % (self.base_dir,self.scene_name)):
        #    with open(g) as f:
        #        j = json.loads(f.read())
        #        pos = j['agent']['position']
        #        key = key_for_point(pos['x'], pos['z'])
        #        pos_id = os.path.splitext(os.path.basename(g))[0]
        #        event_path = os.path.join('%s/%s/events/%s.pickle' % (self.base_dir, self.scene_name, pos_id))
        #        self.positions[key].append({'event': event_path, 'metadata': j})

        # import random
        # total = len(self.positions)

        # p = self.positions[list(self.positions.keys())[random.randint(0, total - 1)]][3]
        # self.last_event = self.load_event(p)

    @property
    def position_x(self):
        return self.last_event.metadata["agent"]["position"]["x"]

    @property
    def position_z(self):
        return self.last_event.metadata["agent"]["position"]["z"]

    @property
    def rotation(self):
        return self.last_event.metadata["agent"]["rotation"]["y"]

    @property
    def camera_horizon(self):
        return self.last_event.metadata["agent"]["cameraHorizon"]

    def start(self):
        pass

    def load_event(self, pos):
        with open(pos["event"], "rb") as f:
            e = pickle.load(f)
        return e

    def find_position(self, x, z, rotation, camera_horizon):
        for p in self.positions.get(ai2thor.controller.key_for_point(x, z), []):
            if (
                abs(p["rotation"] - rotation) < 1.0
                and abs(p["cameraHorizon"] - camera_horizon) < 1.0
            ):
                event = self.load_event(p)
                return event

        return None

    def move(self, x, z):
        return self.find_position(x, z, self.rotation, self.camera_horizon)

    def move_back(self):
        m = MOVE_AHEAD_MAP[self.rotation]
        new_x = (-m["x"] * self.grid_size) + self.position_x
        new_z = (-m["z"] * self.grid_size) + self.position_z
        return self.move(new_x, new_z)

    def move_ahead(self):
        m = MOVE_AHEAD_MAP[self.rotation]
        new_x = (m["x"] * self.grid_size) + self.position_x
        new_z = (m["z"] * self.grid_size) + self.position_z
        return self.move(new_x, new_z)

    def look(self, new_horizon):
        if new_horizon < -30 or new_horizon > 30:
            return None
        else:
            return self.find_position(
                self.position_x, self.position_z, self.rotation, new_horizon
            )

    def look_up(self):
        return self.look(self.camera_horizon - 30)

    def look_down(self):
        return self.look(self.camera_horizon + 30)

    def rotate_right(self):
        new_rotation = (self.rotation + 90) % 360
        return self.rotate(new_rotation)

    def rotate(self, new_rotation):
        return self.find_position(
            self.position_x, self.position_z, new_rotation, self.camera_horizon
        )

    def rotate_left(self):
        new_rotation = (self.rotation - 90) % 360
        return self.rotate(new_rotation)

    def step(self, action=None, **action_args):
        if type(action) is dict:
            action = copy.deepcopy(action)  # prevent changes from leaking
        else:
            action = dict(action=action)

        actions = dict(
            RotateRight=self.rotate_right,
            RotateLeft=self.rotate_left,
            MoveAhead=self.move_ahead,
            MoveBack=self.move_back,
            LookUp=self.look_up,
            LookDown=self.look_down,
        )

        event = actions[action["action"]]()
        if event is None:
            event = copy.deepcopy(self.last_event)
            event.metadata["lastActionSuccess"] = False
            event.metadata["lastAction"] = action["action"]

        self.last_event = event
        return event


class FrameCounter:
    def __init__(self):
        self.counter = 0

    def inc(self):
        self.counter += 1


def write_frame(event, base_dir, scene_name, frame_name):

    events_dir = "%s/%s/events" % (base_dir, scene_name)
    met_dir = "%s/%s/metadata" % (base_dir, scene_name)
    os.makedirs(met_dir, exist_ok=True)
    os.makedirs(events_dir, exist_ok=True)

    with open(events_dir + "/%03d.pickle" % frame_name, "wb") as f:
        pickle.dump(event, f)

    with open(met_dir + "/%03d.json" % frame_name, "w") as f:
        f.write(json.dumps(event.metadata))


def look_up_down_write(controller, base_dir, fc, scene_name):

    fc.inc()
    write_frame(controller.step(action="LookUp"), base_dir, scene_name, fc.counter)
    fc.inc()
    write_frame(controller.step(action="LookDown"), base_dir, scene_name, fc.counter)
    fc.inc()
    write_frame(controller.step(action="LookDown"), base_dir, scene_name, fc.counter)
    controller.step(action="LookUp")


def index_metadata(base_dir, scene_name):
    positions_index = defaultdict(list)
    for g in glob.glob("%s/%s/metadata/*.json" % (base_dir, scene_name)):
        with open(g) as f:
            j = json.loads(f.read())
            agent = j["agent"]
            pos = agent["position"]
            key = ai2thor.controller.key_for_point(pos["x"], pos["z"])
            pos_id = os.path.splitext(os.path.basename(g))[0]
            event_path = os.path.join(
                "%s/%s/events/%s.pickle" % (base_dir, scene_name, pos_id)
            )
            positions_index[key].append(
                {
                    "event": event_path,
                    "rotation": agent["rotation"]["y"],
                    "cameraHorizon": agent["cameraHorizon"],
                }
            )
    with open("%s/%s/index.json" % (base_dir, scene_name), "w") as f:
        f.write(json.dumps(positions_index))


def dump_scene_controller(base_dir, controller):
    if controller.last_event is None:
        raise Exception("Controller must be reset and intialized to a scene")

    scene_name = controller.last_event.metadata["sceneName"]
    fc = FrameCounter()

    shutil.rmtree("%s/%s" % (base_dir, scene_name), ignore_errors=True)

    event = controller.step(action="GetReachablePositions")
    for p in event.metadata["actionReturn"]:
        action = copy.deepcopy(p)
        action["action"] = "TeleportFull"
        action["horizon"] = 0.0
        action["forceAction"] = True
        action["rotation"] = dict(y=0.0)
        event = controller.step(action)
        print(fc.counter)
        if event.metadata["lastActionSuccess"]:
            look_up_down_write(controller, base_dir, fc, scene_name)
            for i in range(3):
                controller.step(action="RotateRight")
                look_up_down_write(controller, base_dir, fc, scene_name)

    index_metadata(base_dir, scene_name)


def dump_scene(
    scene_name,
    base_dir,
    renderInstanceSegmentation=False,
    renderDepthImage=False,
    renderSemanticSegmentation=False,
):
    controller = ai2thor.controller.Controller()
    controller.start(height=448, width=448)
    controller.reset(scene_name)
    controller.step(
        dict(
            action="Initialize",
            fieldOfView=90,
            gridSize=0.25,
            renderDepthImage=renderDepthImage,
            renderInstanceSegmentation=renderInstanceSegmentation,
            renderSemanticSegmentation=renderSemanticSegmentation,
        )
    )
    dump_scene_controller(base_dir, controller)
    controller.stop()



import numpy as np

REAL_2_SIM_TRANSFORM = np.array(
    [[1.00854301, -0.0111386, 0.51920809], [0.00316833, 0.97336625, -1.15532594]]
)


def transform_real_2_sim(real_position):
    """
    Transforms a position from the 'real' coordinate system to the 'sim' coordinate system.
    :param real_position: dictionary with 'x', 'y' and 'z' keys to floating point values
    :return: position in sim space as dictionary with 'x', 'y' and 'z' keys to floating point values
    """

    real_pos = np.array([real_position["x"], real_position["y"], 1])

    sim_pos_np = np.dot(REAL_2_SIM_TRANSFORM, real_pos)

    sim_pos = {"x": sim_pos_np[0], "y": 0.9010001, "z": sim_pos_np[1]}

    return sim_pos



import os
import msgpack
import numpy as np
import requests
import cv2
from pprint import pprint
import shutil
import copy

from ai2thor.server import Event, MultiAgentEvent, DepthFormat
from ai2thor.interact import InteractiveControllerPrompt, DefaultActions


class Controller(object):
    def __init__(
        self,
        headless=False,
        host="127.0.0.1",
        port=0,
        width=300,
        height=300,
        agent_id=0,
        image_dir=".",
        save_image_per_frame=False,
        depth_format=DepthFormat.Meters,
        camera_near_plane=0.1,
        camera_far_plane=20,
    ):
        self.host = host
        self.port = port
        self.headless = headless
        self.last_event = {}
        self.last_action = {}
        self.sequence_id = 0
        self.agent_id = agent_id
        self.screen_width = width
        self.screen_height = height
        self.depth_format = depth_format
        self.camera_near_plane = (camera_near_plane,)
        self.camera_far_plane = camera_far_plane

        if image_dir != ".":
            if os.path.exists(image_dir):
                shutil.rmtree(image_dir)
            os.makedirs(image_dir)

        self.interactive_controller = InteractiveControllerPrompt(
            [
                DefaultActions.MoveAhead,
                DefaultActions.MoveBack,
                DefaultActions.RotateLeft,
                DefaultActions.RotateRight,
                DefaultActions.LookUp,
                DefaultActions.LookDown,
            ],
            image_dir=image_dir,
            image_per_frame=save_image_per_frame,
        )

        self.start(port, host, agent_id=agent_id)

    def stop(self):
        pass

    def start(self, port=9200, host="127.0.0.1", agent_id=0, **kwargs):
        self.host = host
        self.port = port
        self.agent_id = agent_id

        # response_payload = self._post_event('start')
        pprint("-- Start:")
        # pprint(response_payload)

    def reset(self, scene_name=None):
        self.sequence_id = 0
        # response_payload = self._post_event(
        #     'reset', dict(action='Reset', sceneName=scene_name, sequenceId=self.sequence_id)
        # )
        pprint("-- Reset:")
        # pprint(response_payload)

        return self.last_event

    def step(self, action=None, **action_args):

        if type(action) is dict:
            action = copy.deepcopy(action)  # prevent changes from leaking
        else:
            action = dict(action=action)

        raise_for_failure = action_args.pop("raise_for_failure", False)

        action.update(action_args)

        if self.headless:
            action["renderImage"] = False

        action["sequenceId"] = self.sequence_id
        action["agentId"] = self.agent_id

        self.last_action = action

        rotation = action.get("rotation")
        if rotation is not None and type(rotation) != dict:
            action["rotation"] = {}
            action["rotation"]["y"] = rotation

        payload = self._post_event("step", action)
        events = []
        for i, agent_metadata in enumerate(payload["metadata"]["agents"]):
            event = Event(agent_metadata)
            image_mapping = dict(
                image=lambda x: event.add_image(x, flip_y=False, flip_rb_colors=False),
                image_depth=lambda x: event.add_image_depth_robot(
                    x,
                    self.depth_format,
                    camera_near_plane=self.camera_near_plane,
                    camera_far_plane=self.camera_far_plane,
                    flip_y=False,
                    dtype=np.float64,
                ),
            )
            for key in image_mapping.keys():
                if key in payload and len(payload[key]) > i:
                    image_mapping[key](payload[key][i])
            events.append(event)

        if len(events) > 1:
            self.last_event = MultiAgentEvent(self.agent_id, events)
        else:
            self.last_event = events[0]

        if (
            not self.last_event.metadata["lastActionSuccess"]
            and self.last_event.metadata["errorCode"] == "InvalidAction"
        ):
            raise ValueError(self.last_event.metadata["errorMessage"])

        if raise_for_failure:
            assert self.last_event.metadata["lastActionSuccess"]

        # pprint("Display event:")
        # Controller._display_step_event(self.last_event)

        return self.last_event

    def interact(
        self,
        semantic_segmentation_frame=False,
        instance_segmentation_frame=False,
        depth_frame=False,
        color_frame=False,
        metadata=False,
    ):
        self.interactive_controller.interact(
            self,
            semantic_segmentation_frame=semantic_segmentation_frame,
            instance_segmentation_frame=instance_segmentation_frame,
            depth_frame=depth_frame,
            color_frame=color_frame,
            metadata=metadata,
        )

    def _post_event(self, route="", data=None):
        r = requests.post(self._get_url(route), json=data)
        pprint('ACTION "{}"'.format(data["action"]))
        pprint("POST")
        # pprint(r.content)
        pprint(r.status_code)
        return msgpack.unpackb(r.content, raw=False)

    def _get_url(self, route=""):
        return "http://{}:{}{}".format(
            self.host, self.port, "/{}".format(route) if route != "" else ""
        )

    @staticmethod
    def _display_step_event(event):
        metadata = event.metadata
        pprint(metadata)
        cv2.imshow("aoeu", event.cv2img)
        cv2.waitKey(1000)



from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
import os
import requests
import json
from ai2thor.util import makedirs
import ai2thor.downloader
import zipfile
import logging
from ai2thor.util.lock import LockSh, LockEx
from ai2thor.util import atomic_write
import io

logger = logging.getLogger(__name__)

PUBLIC_S3_BUCKET = "ai2-thor-public"
PUBLIC_WEBGL_S3_BUCKET = "ai2-thor-webgl-public"
PRIVATE_S3_BUCKET = "ai2-thor-private"
PYPI_S3_BUCKET = "ai2-thor-pypi"

COMMIT_ID = None
try:
    import ai2thor._builds

    COMMIT_ID = ai2thor._builds.COMMIT_ID
except ImportError:
    pass

platform_map = dict(Linux64="Linux", OSXIntel64="Darwin")
arch_platform_map = {v: k for k, v in platform_map.items()}


def build_name(arch, commit_id, include_private_scenes=False):
    if include_private_scenes:
        return "thor-private-%s-%s" % (arch, commit_id)
    else:
        return "thor-%s-%s" % (arch, commit_id)


def boto_auth():
    return BotoAWSRequestsAuth(
        aws_host="s3-us-west-2.amazonaws.com", aws_region="us-west-2", aws_service="s3"
    )


base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PUBLIC_S3_BUCKET
private_base_url = "http://s3-us-west-2.amazonaws.com/%s/" % PRIVATE_S3_BUCKET


# dummy build when connecting to the editor
class EditorBuild(object):
    def __init__(self):
        # assuming that an external build supports both server types
        self.server_types = ["FIFO", "WSGI"]
        self.url = None
        self.unity_proc = None

    def download(self):
        pass

    def unlock(self):
        pass

    def lock_sh(self):
        pass


class ExternalBuild(object):
    def __init__(self, executable_path):
        self.executable_path = executable_path

        # assuming that an external build supports both server types
        self.server_types = ["FIFO", "WSGI"]

    def download(self):
        pass

    def unlock(self):
        pass

    def lock_sh(self):
        pass


class Build(object):
    def __init__(self, arch, commit_id, include_private_scenes, releases_dir=None):

        self.arch = arch
        self.commit_id = commit_id
        self.include_private_scenes = include_private_scenes
        self.releases_dir = releases_dir
        self.tmp_dir = None

        if self.releases_dir:
            self.tmp_dir = os.path.normpath(self.releases_dir + "/../tmp")

    def download(self):

        makedirs(self.releases_dir)
        makedirs(self.tmp_dir)

        with LockEx(os.path.join(self.tmp_dir, self.name)):
            if not os.path.isdir(self.base_dir):
                z = self.zipfile()
                # use tmpdir instead or a random number
                extract_dir = os.path.join(self.tmp_dir, self.name)
                logger.debug("Extracting zipfile %s" % os.path.basename(self.url))
                z.extractall(extract_dir)

                os.rename(extract_dir, self.base_dir)
                # This can be removed after migrating OSXIntel64 builds to have the AI2Thor executable
                if os.path.exists(self.old_executable_path) and not os.path.exists(
                    self.executable_path
                ):
                    os.link(self.old_executable_path, self.executable_path)

                # we can lose the executable permission when unzipping a build
                os.chmod(self.executable_path, 0o755)

            else:
                logger.debug("%s exists - skipping download" % (self.executable_path,))

    def zipfile(self):
        zip_data = ai2thor.downloader.download(
            self.url, self.sha256(), self.include_private_scenes
        )

        return zipfile.ZipFile(io.BytesIO(zip_data))

    def download_metadata(self):
        # this can happen if someone has an existing release without the metadata
        # built prior to the backfill
        # can add check to see if metadata has expired/we update metadata
        # if we want to add more info to metadata
        if not self.metadata:
            z = self.zipfile()
            atomic_write(self.metadata_path, z.read("metadata.json"))

    @property
    def old_executable_path(self):
        if self.arch == "Linux64":
            return self.executable_path
        elif self.arch == "OSXIntel64":
            return os.path.join(
                self.base_dir, self.name + ".app", "Contents/MacOS", self.name
            )
        else:
            raise Exception("unable to handle target arch %s" % self.arch)

    def parse_plist(self):
        import xml.etree.ElementTree as ET

        plist_path = os.path.join(
            self.base_dir, self.name + ".app", "Contents/Info.plist"
        )

        with open(plist_path) as f:
            plist = f.read()
        root = ET.fromstring(plist)

        keys = [x.text for x in root.findall("dict/key")]
        values = [x.text for x in root.findall("dict/string")]
        return dict(zip(keys, values))

    @property
    def base_dir(self):
        return os.path.join(self.releases_dir, self.name)

    @property
    def executable_path(self):
        if self.arch == "Linux64":
            return os.path.join(self.base_dir, self.name)
        elif self.arch == "OSXIntel64":
            plist = self.parse_plist()
            return os.path.join(
                self.base_dir,
                self.name + ".app",
                "Contents/MacOS",
                plist["CFBundleExecutable"],
            )
        else:
            raise Exception("unable to handle target arch %s" % self.arch)

    @property
    def metadata_path(self):
        return os.path.join(self.base_dir, "metadata.json")

    @property
    def metadata(self):
        if os.path.isfile(self.metadata_path):
            with open(self.metadata_path, "r") as f:
                return json.loads(f.read())
        else:
            return None

    @property
    def server_types(self):
        self.download_metadata()
        return self.metadata.get("server_types", [])

    def auth(self):
        if self.include_private_scenes:
            return boto_auth()
        else:
            return None

    def _base_url(self):
        if self.include_private_scenes:
            return private_base_url
        else:
            return base_url

    @property
    def url(self):
        return self._base_url() + os.path.join("builds", self.name + ".zip")

    @property
    def name(self):
        return build_name(self.arch, self.commit_id, self.include_private_scenes)

    def unlock(self):
        if self._lock:
            self._lock.unlock()
            self._lock = None

    def lock_sh(self):
        self._lock = LockSh(self.base_dir)
        self._lock.lock()

    @property
    def log_url(self):
        return os.path.splitext(self.url)[0] + ".log"

    @property
    def metadata_url(self):
        return os.path.splitext(self.url)[0] + ".json"

    @property
    def sha256_url(self):
        return os.path.splitext(self.url)[0] + ".sha256"

    def exists(self):
        return (self.releases_dir and os.path.isdir(self.base_dir)) or (
            requests.head(self.url, auth=self.auth()).status_code == 200
        )

    def log_exists(self):
        return requests.head(self.log_url, auth=self.auth()).status_code == 200

    def sha256(self):
        res = requests.get(self.sha256_url, auth=self.auth())
        res.raise_for_status()
        return res.content.decode("ascii")



from pprint import pprint
from unityparser import UnityDocument

# Requires unity parser, run:
# pip install unityparser


def updateNavMeshParamsForScene(scene_file_name):
    pprint("Updating file '{}'...".format(scene_file_name))
    doc = UnityDocument.load_yaml(scene_file_name)
    for entry in doc.entries:
        if entry.__class__.__name__ == "NavMeshSettings":
            # print(entry.__class__.__name__)
            buildSettings = getattr(entry, "m_BuildSettings", None)
            # pprint(buildSettings)
            buildSettings["agentRadius"] = "0.2"
            buildSettings["agentHeight"] = "1.8"
            buildSettings["agentClimb"] = "0.5"
            buildSettings["manualCellSize"] = "1"

            buildSettings["cellSize"] = "0.03"

    doc.dump_yaml()


def GetRoboSceneNames(
    last_index, last_subIndex, nameTemplate, prefix_path="unity/Assets/Scenes"
):
    return [
        "{}/FloorPlan_{}{}_{}.unity".format(prefix_path, nameTemplate, i, j)
        for i in range(1, last_index + 1)
        for j in range(1, last_subIndex + 1)
    ]


def GetSceneNames(
    start_index, last_index, nameTemplate="", prefix_path="unity/Assets/Scenes"
):
    return [
        "{}/FloorPlan{}{}_physics.unity".format(prefix_path, i, nameTemplate)
        for i in range(start_index, last_index + 1)
    ]


def main():

    # testSceneNames = GetRoboSceneNames(3, 5, "Val")
    # valSceneNames = GetRoboSceneNames(2, 2, "test-dev", "unity/Assets/Private/Scenes")
    # trainSceneNames = GetRoboSceneNames(12, 5, "Train")
    # allScenes = testSceneNames  + trainSceneNames
    # allScenes = valSceneNames

    iThorScenes = (
        GetSceneNames(1, 30)
        + GetSceneNames(201, 230)
        + GetSceneNames(301, 330)
        + GetSceneNames(401, 430)
        + GetSceneNames(501, 530)
    )
    allScenes = iThorScenes
    # print(allScenes)
    for scene_file_name in allScenes:
        updateNavMeshParamsForScene(scene_file_name)
        # print(scene_file_name)


if __name__ == "__main__":
    main()

    # Exceptions:
    # Scene FloorPlan_Train7_1
    # Train_11_3 unmade bed
    # Val2_3 unamde bed



# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.server

Handles all communication with Unity through a Flask service.  Messages
are sent to the controller using a pair of request/response queues.
"""
import ai2thor.server
import json
import logging
import threading
import os

try:
    from queue import Queue, Empty
except ImportError:
    from Queue import Queue, Empty

import time

from flask import Flask, request, make_response, abort
import werkzeug
import werkzeug.serving
import werkzeug.http

logging.getLogger("werkzeug").setLevel(logging.ERROR)

werkzeug.serving.WSGIRequestHandler.protocol_version = "HTTP/1.1"

# get with timeout to allow quit
def queue_get(que, unity_proc=None):
    res = None
    attempts = 0
    max_attempts = 200
    while True:
        try:
            res = que.get(block=True, timeout=0.5)
            break
        except Empty:
            attempts += 1

            # we poll here for the unity proc in the event that it has
            # exited otherwise we would wait indefinetly for the queue
            if unity_proc:
                if unity_proc.poll() is not None:
                    raise Exception("Unity process exited %s" % unity_proc.returncode)

                # no Action should take > 100s to complete, so we assume that
                # something has gone wrong within Unity
                # max_attempts can also be triggered if an Exception is thrown from
                # within the thread used to run the wsgi server, in which case
                # Unity will receive a corrupted response
                if attempts >= max_attempts:
                    raise Exception(
                        "Could not get a message from the queue after %s attempts "
                        % attempts
                    )
    return res


class BufferedIO(object):
    def __init__(self, wfile):
        self.wfile = wfile
        self.data = []

    def write(self, output):
        self.data.append(output)

    def flush(self):
        self.wfile.write(b"".join(self.data))
        self.wfile.flush()

    def close(self):
        return self.wfile.close()

    @property
    def closed(self):
        return self.wfile.closed


class ThorRequestHandler(werkzeug.serving.WSGIRequestHandler):
    def run_wsgi(self):
        old_wfile = self.wfile
        self.wfile = BufferedIO(self.wfile)
        result = super(ThorRequestHandler, self).run_wsgi()
        self.wfile = old_wfile
        return result


class MultipartFormParser(object):
    @staticmethod
    def get_boundary(request_headers):
        for h, value in request_headers:
            if h == "Content-Type":
                ctype, ct_opts = werkzeug.http.parse_options_header(value)
                boundary = ct_opts["boundary"].encode("ascii")
                return boundary
        return None

    def __init__(self, data, boundary):

        self.form = {}
        self.files = {}

        full_boundary = b"--" + boundary
        mid_boundary = b"\r\n" + full_boundary
        view = memoryview(data)
        i = data.find(full_boundary) + len(full_boundary)
        while i >= 0:
            next_offset = data.find(mid_boundary, i)
            if next_offset < 0:
                break
            headers_offset = i + 2  # add 2 for CRLF
            body_offset = data.find(b"\r\n\r\n", headers_offset)
            raw_headers = view[headers_offset:body_offset]
            body = view[body_offset + 4 : next_offset]
            i = next_offset + len(mid_boundary)

            headers = {}
            for header in raw_headers.tobytes().decode("ascii").strip().split("\r\n"):

                k, v = header.split(":")
                headers[k.strip()] = v.strip()

            ctype, ct_opts = werkzeug.http.parse_options_header(headers["Content-Type"])
            cdisp, cd_opts = werkzeug.http.parse_options_header(
                headers["Content-disposition"]
            )
            assert cdisp == "form-data"

            if "filename" in cd_opts:
                if cd_opts["name"] not in self.files:
                    self.files[cd_opts["name"]] = []

                self.files[cd_opts["name"]].append(body)

            else:
                if ctype == "text/plain" and "charset" in ct_opts:
                    body = body.tobytes().decode(ct_opts["charset"])
                if cd_opts["name"] not in self.form:
                    self.form[cd_opts["name"]] = []

                self.form[cd_opts["name"]].append(body)


class WsgiServer(ai2thor.server.Server):

    server_type = "WSGI"

    def __init__(
        self,
        host,
        port=0,
        threaded=False,
        depth_format=ai2thor.server.DepthFormat.Meters,
        add_depth_noise=False,
        width=300,
        height=300,
    ):

        app = Flask(
            __name__,
            template_folder=os.path.realpath(
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "..", "templates"
                )
            ),
        )

        self.request_queue = Queue(maxsize=1)
        self.response_queue = Queue(maxsize=1)
        self.app = app
        self.app.config.update(
            PROPAGATE_EXCEPTIONS=False, JSONIFY_PRETTYPRINT_REGULAR=False
        )
        self.port = port
        self.last_rate_timestamp = time.time()
        self.frame_counter = 0
        self.debug_frames_per_interval = 50
        self.unity_proc = None
        self.wsgi_server = werkzeug.serving.make_server(
            host,
            self.port,
            self.app,
            threaded=threaded,
            request_handler=ThorRequestHandler,
        )
        # used to ensure that we are receiving frames for the action we sent
        super().__init__(width, height, depth_format, add_depth_noise)

        @app.route("/ping", methods=["get"])
        def ping():
            return "pong"

        @app.route("/train", methods=["post"])
        def train():

            action_returns = []

            if request.headers["Content-Type"].split(";")[0] == "multipart/form-data":
                form = MultipartFormParser(
                    request.get_data(),
                    MultipartFormParser.get_boundary(request.headers),
                )
                metadata = json.loads(form.form["metadata"][0])
                # backwards compatibility
                if (
                    "actionReturns" in form.form
                    and len(form.form["actionReturns"][0]) > 0
                ):
                    action_returns = json.loads(form.form["actionReturns"][0])
                token = form.form["token"][0]
            else:
                form = request
                metadata = json.loads(form.form["metadata"])
                # backwards compatibility
                if "actionReturns" in form.form and len(form.form["actionReturns"]) > 0:
                    action_returns = json.loads(form.form["actionReturns"])
                token = form.form["token"]

            if self.client_token and token != self.client_token:
                abort(403)

            if self.frame_counter % self.debug_frames_per_interval == 0:
                now = time.time()
                # rate = self.debug_frames_per_interval / float(now - self.last_rate_timestamp)
                self.last_rate_timestamp = now
                # import datetime
                # print("%s %s/s" % (datetime.datetime.now().isoformat(), rate))

            for i, a in enumerate(metadata["agents"]):
                if "actionReturn" not in a and i < len(action_returns):
                    a["actionReturn"] = action_returns[i]

            event = self.create_event(metadata, form.files)

            self.request_queue.put_nowait(event)

            self.frame_counter += 1

            next_action = queue_get(self.response_queue)
            if "sequenceId" not in next_action:
                self.sequence_id += 1
                next_action["sequenceId"] = self.sequence_id
            else:
                self.sequence_id = next_action["sequenceId"]

            resp = make_response(
                json.dumps(next_action, cls=ai2thor.server.NumpyAwareEncoder)
            )

            return resp

    def _start_server_thread(self):
        self.wsgi_server.serve_forever()

    def start(self):
        self.started = True
        self.server_thread = threading.Thread(target=self._start_server_thread)
        self.server_thread.daemon = True
        self.server_thread.start()

    def receive(self):
        return queue_get(self.request_queue, unity_proc=self.unity_proc)

    def send(self, action):
        assert self.request_queue.empty()
        self.response_queue.put_nowait(action)

    # params to pass up to unity
    def unity_params(self):
        host, port = self.wsgi_server.socket.getsockname()

        params = dict(host=host, port=str(port))
        return params

    def stop(self):
        self.send({})
        self.wsgi_server.shutdown()



from setuptools import setup, find_packages
import os

__version__ = "0.0.1"

if os.path.isfile("ai2thor/_version.py"):
    exec(open("ai2thor/_version.py").read())

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

VERSION = __version__

setup(
    name="ai2thor",
    version=VERSION,
    description="AI2-THOR: A Near Photo-Realistic Interactable Framework for Embodied AI Agents",
    long_description=long_description,
    license="Apache License 2.0",
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    keywords="AI2-THOR, Allen AI, Python, Reinforcement Learning, Computer Vision, Artificial Intelligence",
    url="https://github.com/allenai/ai2thor",
    author="Allen Institute for AI",
    author_email="ai2thor@allenai.org",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        "flask",
        "numpy",
        "pyyaml",
        "requests",
        "progressbar2",
        "botocore",
        "aws-requests-auth",
        "msgpack",
        "Pillow",
        "opencv-python",
        "werkzeug>=0.15.0",  # needed for unix socket support
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-timeout", "pytest-cov", "jsonschema", "shapely"],
    scripts=["scripts/ai2thor-xorg"],
    include_package_data=False,
)



# Copyright Allen Institute for Artificial Intelligence 2017
"""
ai2thor.server

Handles all communication with Unity through a Flask service.  Messages
are sent to the controller using a pair of request/response queues.
"""
import ai2thor.server
import json
import msgpack
import os
import tempfile
from ai2thor.exceptions import UnityCrashException
from enum import IntEnum, unique
from collections import defaultdict
import struct

# FifoFields
@unique
class FieldType(IntEnum):
    METADATA = 1
    ACTION = 2
    ACTION_RESULT = 3
    RGB_IMAGE = 4
    DEPTH_IMAGE = 5
    NORMALS_IMAGE = 6
    FLOWS_IMAGE = 7
    CLASSES_IMAGE = 8
    IDS_IMAGE = 9
    THIRD_PARTY_IMAGE = 10
    METADATA_PATCH = 11
    THIRD_PARTY_DEPTH = 12
    THIRD_PARTY_NORMALS = 13
    THIRD_PARTY_IMAGE_IDS = 14
    THIRD_PARTY_CLASSES = 15
    THIRD_PARTY_FLOW = 16
    END_OF_MESSAGE = 255


class FifoServer(ai2thor.server.Server):
    header_format = "!BI"
    header_size = struct.calcsize(header_format)
    field_types = {f.value: f for f in FieldType}
    server_type = "FIFO"

    def __init__(
        self,
        width,
        height,
        depth_format=ai2thor.server.DepthFormat.Meters,
        add_depth_noise=False,
    ):

        self.tmp_dir = tempfile.TemporaryDirectory()
        self.server_pipe_path = os.path.join(self.tmp_dir.name, "server.pipe")
        self.client_pipe_path = os.path.join(self.tmp_dir.name, "client.pipe")
        self.server_pipe = None
        self.client_pipe = None
        self.raw_metadata = None
        self.raw_files = None
        self._last_action_message = None
        # allows us to map the enum to form field names
        # for backwards compatibility
        # this can be removed when the wsgi server is removed
        self.form_field_map = {
            FieldType.RGB_IMAGE: "image",
            FieldType.DEPTH_IMAGE: "image_depth",
            FieldType.CLASSES_IMAGE: "image_classes",
            FieldType.IDS_IMAGE: "image_ids",
            FieldType.NORMALS_IMAGE: "image_normals",
            FieldType.FLOWS_IMAGE: "image_flow",
            FieldType.THIRD_PARTY_IMAGE: "image-thirdParty-camera",
            FieldType.THIRD_PARTY_DEPTH: "image_thirdParty_depth",
            FieldType.THIRD_PARTY_NORMALS: "image_thirdParty_normals",
            FieldType.THIRD_PARTY_IMAGE_IDS: "image_thirdParty_image_ids",
            FieldType.THIRD_PARTY_CLASSES: "image_thirdParty_classes",
            FieldType.THIRD_PARTY_FLOW: "image_thirdParty_flow",
        }

        self.image_fields = {
            FieldType.IDS_IMAGE,
            FieldType.CLASSES_IMAGE,
            FieldType.FLOWS_IMAGE,
            FieldType.NORMALS_IMAGE,
            FieldType.DEPTH_IMAGE,
            FieldType.RGB_IMAGE,
            FieldType.THIRD_PARTY_IMAGE,
            FieldType.THIRD_PARTY_DEPTH,
            FieldType.THIRD_PARTY_NORMALS,
            FieldType.THIRD_PARTY_IMAGE_IDS,
            FieldType.THIRD_PARTY_CLASSES,
            FieldType.THIRD_PARTY_FLOW,
        }

        self.eom_header = self._create_header(FieldType.END_OF_MESSAGE, b"")
        super().__init__(width, height, depth_format, add_depth_noise)

    def _create_header(self, message_type, body):
        return struct.pack(self.header_format, message_type, len(body))

    def _recv_message(self):
        if self.server_pipe is None:
            self.server_pipe = open(self.server_pipe_path, "rb")

        metadata = None
        files = defaultdict(list)
        while True:
            header = self.server_pipe.read(self.header_size)  # message type + length
            if len(header) == 0:
                self.unity_proc.wait(timeout=5)
                returncode = self.unity_proc.returncode
                message = (
                    "Unity process has exited - check Player.log for errors. Last action message: %s, returncode=%s"
                    % (self._last_action_message, self.unity_proc.returncode)
                )
                # we don't want to restart all process exits since its possible that a user
                # kills off a Unity process with SIGTERM to end a training run
                # SIGABRT is the returncode for when Unity crashes due to a segfault
                if returncode in [-6, -11]:  # SIGABRT, SIGSEGV
                    raise UnityCrashException(message)
                else:
                    raise Exception(message)

            if header[0] == FieldType.END_OF_MESSAGE.value:
                # print("GOT EOM")
                break

            # print("got header %s" % header)
            field_type_int, message_length = struct.unpack(self.header_format, header)
            field_type = self.field_types[field_type_int]
            body = self.server_pipe.read(message_length)
            # print("field type")
            # print(field_type)
            if field_type is FieldType.METADATA:
                # print("body length %s" % len(body))
                # print(body)
                metadata = msgpack.loads(body, raw=False)
            elif field_type is FieldType.METADATA_PATCH:
                metadata_patch = msgpack.loads(body, raw=False)
                agents = self.raw_metadata["agents"]
                metadata = dict(
                    agents=[{} for i in range(len(agents))],
                    thirdPartyCameras=self.raw_metadata["thirdPartyCameras"],
                    sequenceId=self.sequence_id,
                    activeAgentId=metadata_patch["agentId"],
                )
                for i in range(len(agents)):
                    metadata["agents"][i].update(agents[i])

                metadata["agents"][metadata_patch["agentId"]].update(metadata_patch)
                files = self.raw_files
            elif field_type in self.image_fields:
                files[self.form_field_map[field_type]].append(body)
            else:
                raise ValueError("Invalid field type: %s" % field_type)

        self.raw_metadata = metadata
        self.raw_files = files

        return metadata, files

    def _send_message(self, message_type, body):
        # print("trying to write to ")
        if self.client_pipe is None:
            self.client_pipe = open(self.client_pipe_path, "wb")

        header = self._create_header(message_type, body)
        # print("len header %s" % len(header))
        # print("sending body %s" % body)

        # used for debugging in case of an error
        self._last_action_message = body

        self.client_pipe.write(header + body + self.eom_header)
        self.client_pipe.flush()

    def receive(self):

        metadata, files = self._recv_message()

        if metadata is None:
            raise ValueError("no metadata received from recv_message")

        return self.create_event(metadata, files)

    def send(self, action):
        # print("got action to send")
        if "sequenceId" in action:
            self.sequence_id = action["sequenceId"]
        else:
            self.sequence_id += 1
            action["sequenceId"] = self.sequence_id
        # print(action)

        # need to switch this to msgpack
        self._send_message(
            FieldType.ACTION,
            json.dumps(action, cls=ai2thor.server.NumpyAwareEncoder).encode("utf8"),
        )

    def start(self):
        os.mkfifo(self.server_pipe_path)
        os.mkfifo(self.client_pipe_path)
        self.started = True

    # params to pass up to unity
    def unity_params(self):
        params = dict(
            fifo_server_pipe_path=self.server_pipe_path,
            fifo_client_pipe_path=self.client_pipe_path,
        )
        return params

    def stop(self):
        self.client_pipe.close()
        self.server_pipe.close()
